#include <Imlib2.h>
#include <stdio.h>

#include "imgprint.h"

#define WARN	"Warning: "

//#define dprintf printf
#define dprintf(...)

static int avg(DATA32 pix) {
	int sum = pix & 0xFF;
	sum += (pix >> 8) & 0xFF;
	sum += (pix >> 16) & 0xFF;
	return sum / 3;
}

static int slika_init(FILE *fo, int w, int h) {
	const unsigned char ini[] = {033, 'K', (w >> 3) + !!(w & 7), h & 0xFF, h >> 8};

	if (!w || !h) {
		fprintf(stderr, "Slika je premala (w|h < 0)!\n");
		return 1;
	}

	if (w > MAX_W) {
		fprintf(stderr, "Slika je preširoka!\n");
		return 1;
	}

	if (h > MAX_H) {
		fprintf(stderr, "Slika je previsoka!\n");
		return 1;
	}

	fwrite(ini, sizeof(ini), 1, fo);

	return 0;
}

int ispisi_sliku(const unsigned long *img, int w, int h, FILE *fo) {
	int x, y;
	
	if (slika_init(fo, w, h))
		return 1;

	for (y = 0; y < h; ++y) {
		unsigned char z;
		for (x = 0; x < w; ++x) {
			char p = !(avg(img[y * w + x]) & 0x80);
			dprintf("%c", p + '0');
			z = (z << 1) | p;
			if ((x & 7) == 7) {
//				dprintf("|%02X|", z);
				fwrite(&z, 1, 1, fo);
			}
		}
		if (x & 7) {
			z <<= 8 - (x & 7);
//			dprintf("|%02X|", z);
			fwrite(&z, 1, 1, fo);
		}
		dprintf("\n");
	}

	return 0;
}

int ucitaj_i_ispisi_sliku(const char *fnin, FILE *fo) {
	Imlib_Image image, tmp;
	int w, h, rv;

	image = imlib_load_image(fnin);
	if (!image) {
		fprintf(stderr, "Ne mogu učitati sliku!\n");
		return 1;
	}

	imlib_context_set_image(image);

	w = imlib_image_get_width();
	h = imlib_image_get_height();

	// Preširoka? Rotiraj!
	if (w > MAX_W && w > h) {
		int t;
		imlib_image_orientate(1);
		t = w;
		w = h;
		h = t;

		fprintf(stderr, WARN "Rotirana!\n");
	}

	// Preširoka i dalje?
	if (w > MAX_W) {
		int modif = (w / MAX_W) + 1;

		tmp = imlib_create_cropped_scaled_image(0, 0, w, h, w / modif, h / modif);
		imlib_free_image();
		imlib_context_set_image(tmp);

		w /= modif;
		h /= modif;

		fprintf(stderr, WARN "Resized to: %dx%d\n", w, h);
	}

	rv = ispisi_sliku((const unsigned long *)imlib_image_get_data_for_reading_only(), w, h, fo);

	imlib_free_image();

	return rv;
}
