#include <stdio.h>
#include <getopt.h>
#include <string.h>

#include <string>
#include <map>

#include "imgprint.h"

#define WARN		"Warning: "

#define FONT_OFFSET	0xF0
#define CHAR_WIDTH	7
#define CHAR_HEIGHT	8

#define WIDTH(bold)	((bold) ? 20 : 40)

using namespace std;

typedef struct {
	unsigned int bold	: 1;
	unsigned int align	: 2;
	unsigned int _unused0	: 5;
	unsigned int linespace	: 8;
	unsigned int _unused	: 16;
} format;
#define INIT_FORMAT	((format){0, 0, 0, 2, 0})

static const char CHAR_SO = 0x0E;
static const char CHAR_DC4 = 0x14;
static const char CHAR_CAN = 0x18;

map<const string, string> str_repl_list;
map<int, string> chars_list;

static void send_font(FILE *fo) {
	char ini[4] = {033, '&'};

	for (map<int, string>::const_iterator i = chars_list.begin(); i != chars_list.end(); ++i) {
		if (i->second.size() != CHAR_WIDTH)
			continue;
		ini[2] = ini[3] = i->first + FONT_OFFSET;
		fwrite(ini, sizeof(ini), 1, fo);
		fwrite(i->second.data(), i->second.size(), 1, fo);
		fflush(fo);
	}
}

static void clear_font(int from, int to) {
	if (from < 0 || from > 7 || to < 0 || to > 7 || from > to)
		return;
	for (; from <= to; ++from) {
		chars_list.erase(from);
		map<const string, string>::iterator j, i;
		for (i = str_repl_list.begin(); i != str_repl_list.end();) {
			if (i->second == string(1, FONT_OFFSET + from)) {
				j = i++;
				str_repl_list.erase(j);
				continue;
			}
			++i;
		}
	}
}

static void rmnl(char *s) {
	char *t;
	if ((t = strchr(s, '\n'))) *t = 0;
	if ((t = strchr(s, '\r'))) *t = 0;
}

static string load_char(FILE *f) {
	char red[8][32];
	int x, y;
	string s;
	for (y = 0; y < 8; ++y) {
		if (!fgets(red[y], sizeof(red[0]), f))
			return string();
		rmnl(red[y]);
		if (strlen(red[y]) < 8) {
			int i;
			char *t = red[y] + strlen(red[y]);
			for (i = 0; i < 8; ++i, ++t)
				*t = ' ';
			*t = 0;
		}
	}
	for (x = 0; x < CHAR_WIDTH; ++x) {
		unsigned char z = 0;
		for (y = CHAR_HEIGHT - 1; y >= 0; --y)
			z = (z << 1) | (red[y][x] != ' ');
		s += string(1, z);
	}
	return s;
}

static int load_font(const char *fname) {
	FILE *f;
	char buf[256];
	int ret = 0;

	f = fopen(fname, "rt");
	if (!f) {
		fprintf(stderr, "Ne mogu otvoriti fajl s fontom!\n");
		return 1;
	}

	while (fgets(buf, sizeof(buf), f)) {
		rmnl(buf);
		if (*buf == ':') {
			char id;
			long ofs;
			if (chars_list.size() >= 8) {
				fprintf(stderr, WARN "Svih 8 znakova je iskorišteno!\n");
				break;
			}
			for (id = 0; ; ++id) {
				if (chars_list[id].empty())
					break;
			}

			do {
				str_repl_list[buf+1] = string(1,id+FONT_OFFSET);
				ofs = ftell(f);
				if (!fgets(buf, sizeof(buf), f))
				       break;
				rmnl(buf);
			} while (*buf == ':');
			fseek(f, ofs, SEEK_SET);
			chars_list[id] = load_char(f);
		} else if (*buf == '@') {
			int id;
			if (sscanf(buf+1, "%d", &id) != 1 || id < 0 || id >= 8)
				continue;
			chars_list[id] = load_char(f);
		}
	}

	fclose(f);
	return ret;
}

static void str_replace(string& where, const string& what, const string& by) {
	for (string::size_type i  = where.find(what);
			i != string::npos;
			i  = where.find(what, i + by.size()))
		where.replace(i, what.size(), by);
}

static void all_replace(string &s) {
	map<const string, string>::const_iterator i;
	for (i = str_repl_list.begin(); i != str_repl_list.end(); ++i)
		str_replace(s, i->first, i->second);
}

static int reducegetoct(const char *s) {
	int i;
	if (sscanf(s, "%3o", &i) != 1)
		return -1;
	return i;
}

static int reducegethex(const char *s) {
	int i;
	if (sscanf(s, "x%2x", &i) != 1)
		return -1;
	return i;
}

static void reduce(char *s) {
	int i;

	for (; *s; ++s) {
		if (*s == '\\') {
			switch (s[1]) {
				case 'n':
					*s = '\n';
					break;
				case 'r':
					*s = '\r';
					break;
				case 't':
					*s = '\t';
					break;
				default:
					i = reducegethex(s+1);
					if (1 < 0) i = reducegetoct(s+1);

					if (i >= 0) {
						*s = i;
						memmove(s+2, s+4, strlen(s+2)-1);
					} else {
						goto dalje;
					}
					break;
			}

			memmove(s+1, s+2, strlen(s+1));
		}

dalje:;
	}
}

static void refresh_format(const format &fmt, FILE *fo, bool force = false) {
	static format ofmt = INIT_FORMAT;

	if (fmt.bold)
		fwrite(&CHAR_SO, sizeof(CHAR_SO), 1, fo);

	if (ofmt.linespace != fmt.linespace || force) {
		unsigned char ls[] = {033, 'A', fmt.linespace};
		ofmt.linespace = fmt.linespace;
		fwrite(ls, sizeof(ls), 1, fo);
	}
}

static void paper_feed(FILE *fo, int n) {
	const unsigned char d[] = {033, 'B', n};
	fwrite(d, sizeof(d), 1, fo);
	fflush(fo);
}

static void test_printer(FILE *fo) {
	int i;
	char c;
	for (i = ' '; i <= 0xFF; ++i) {
		c = i;
		fwrite(&c, 1, 1, fo);
	}
	c = '\r';
	fwrite(&c, 1, 1, fo);
	fflush(fo);
}

static size_t realsize(const string &s) {
	int i;
	for (i = s.size() - 1; i >= 0 && (s[i] == '\r' || s[i] == '\n'); --i);
	return (size_t)(i + 1);
}

static void horizontal_rule(FILE *fo, int thick, int w) {
	const int crta = 23 * w / 100, blank = (23 - crta) / 2;
	const unsigned char graph[] = {033, 'K', blank + crta, thick, 0};
	const unsigned char cb = 0, cc = 0xFF;
	int i, j;
	fwrite(graph, sizeof(graph), 1, fo);
	for (j = 0; j < thick; ++j) {
		for (i = 0; i < blank; ++i)
			fwrite(&cb, 1, 1, fo);
		for (i = 0; i < crta; ++i)
			fwrite(&cc, 1, 1, fo);
	}
	fflush(fo);
}

static void main_loop(FILE *fo) {
	char red[256];
	format fmt = INIT_FORMAT;

	while (fgets(red, sizeof(red), stdin)) {
		if (*red == '.') {
			char *cmd = red + 1;
			if (*cmd == '.') {
				memmove(red, red+1, strlen(red));
				goto nast;
			}

			// procesiranje komande...

			rmnl(cmd);
			if (!strcmp(cmd, "b")) {
				fmt.bold = 1;
			} else if (!strcmp(cmd, "nb")) {
				fmt.bold = 0;
			} else if (!strcmp(cmd, "c")) {
				fmt.align = 1;
			} else if (!strcmp(cmd, "l")) {
				fmt.align = 0;
			} else if (!strcmp(cmd, "r")) {
				fmt.align = 2;
			} else if (!strcmp(cmd, "nc") || !strcmp(cmd, "nr")) {
				fmt.align = 0;
			} else if (!strncmp(cmd, "hr ", 3)) {
				int t = 1, w = 100;
				sscanf(cmd + 3, "%d%d", &t, &w);
				if (t > 0 && w > 0)
					horizontal_rule(fo, t, w);
			} else if (!strncmp(cmd, "ls ", 3)) {
				int ls = fmt.linespace;
				sscanf(cmd + 3, "%d", &ls);
				if (ls >= 0 && ls <= 255) {
					fmt.linespace = ls;
					if (ls & 1)
						fprintf(stderr, WARN "Odd linespace!\n");
				}
			} else if (!strncmp(cmd, "pf ", 3)) {
				int pf;
				if (sscanf(cmd + 3, "%d", &pf) && pf >= 4 && pf <= 255)
					paper_feed(fo, pf);
			} else if (!strncmp(cmd, "cfont ", 6)) {
				int from, to;
				if (sscanf(cmd + 6, "%d%d", &from, &to) == 2)
					clear_font(from, to);
			} else if (!strncmp(cmd, "font ", 5)) {
				if (!load_font(cmd + 5))
					send_font(fo);
			} else if (!strncmp(cmd, "img ", 4)) {
				ucitaj_i_ispisi_sliku(cmd + 4, fo);
				fflush(fo);
			} else if (!strcmp(cmd, "test")) {
				test_printer(fo);
			} else if (!strcmp(cmd, "exit") || !strcmp(cmd, "quit")) {
				return;
			} else if (*cmd != '#') {
				fprintf(stderr, "Unknown command: \"%s\"\n", cmd);
			}
			continue;
		}
nast:
		string s = red;
		all_replace(s);
		str_replace(s, "\n", "\r");
		refresh_format(fmt, fo);
		if (fmt.align == 1) {
			int al = (WIDTH(fmt.bold) - realsize(s)) / 2;
			if (al > 0) s = string(al, ' ') + s;
		} else if (fmt.align == 2) {
			int al = WIDTH(fmt.bold) - realsize(s);
			if (al > 0) s = string(al, ' ') + s;
		}
		fwrite(s.data(), s.size(), 1, fo);
		fflush(fo);
	}
}

static void help(const char *pn) {
	fprintf(stderr, "Usage: %s [options]\n", pn);
	fputs(	"Options:\n"
		" --font|-f <fontfile> - font\n"
		" --no-send-font - don't send font\n"
		" --replace|-r <h>=<n> - replace <h> with <n>\n"
		" -e - \\ supported in replace\n"
		" --help|-h - print this help and exit\n"
		"\n"
		"Formatting:\n"
		" .b - bold\n"
		" .nb - normal\n"
		" .c - center\n"
		" .l - left\n"
		" .r - right\n"
		" .hr <thick> <width> - horizontal rule, width in %\n"
		" .ls <ls> - linespace\n"
		" .font <file> - učita font\n"
		" .cfont <from> <to> - izbriše učitane znakove\n"
		" .img <file> - ubaci sliku\n"
		" .# - komentar\n"
		, stderr);
}

int main(int argn, char *argv[]) {
	static const char mini_opts[] = "f:r:eh";
	static struct option long_opts[] = {
		{"font", 1, NULL, 'f'},
		{"replace", 1, NULL, 'r'},
		{"no-send-font", 0, NULL, 1},
		{"no-printer-init", 0, NULL, 2},
		{"help", 0, NULL, 'h'},
		{NULL, 0, NULL, 0}
	};

	int o;
	int oid;
	FILE *fo = fopen("/dev/stdout", "wb");
	struct {
		int extended : 1;
		int sendfont : 1;
		int initprinter : 1;
	} opt = {0, 1, 1};

	while ((o = getopt_long(argn, argv, mini_opts, long_opts, &oid)) != -1)
	switch (o) {
		case 1:
			opt.sendfont = 0;
			break;

		case 2:
			opt.initprinter = 0;
			break;

		case 'e':
			opt.extended = !opt.extended;
			break;

		case 'r':
			char t1[256], t2[256];
			if (sscanf(optarg, "%255[^=]=%255s", t1, t2) != 2) {
				goto arg_error;
			}
			if (opt.extended) {
				reduce(t1);
				reduce(t2);
			}
			str_repl_list[t1] = t2;
			break;

		case 'f':
			if (load_font(optarg))
				return 1;
			break;

		case 'h':
			help(*argv);
			return 0;

		arg_error:
		default:
			help(*argv);
			return 1;
	}

	if (opt.sendfont) {
		send_font(fo);
	}

	if (opt.initprinter) {
		fwrite(&CHAR_DC4, sizeof(CHAR_DC4), 1, fo);
		fwrite(&CHAR_CAN, sizeof(CHAR_CAN), 1, fo);
		refresh_format(INIT_FORMAT, fo, true);
	}

	main_loop(fo);

	return 0;
}
